/**********
This library is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the
Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version. (See <http://www.gnu.org/copyleft/lesser.html>.)

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
more details.

You should have received a copy of the GNU Lesser General Public License
along with this library; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
**********/
// "liveMedia"
// Copyright (c) 1996-2011 Live Networks, Inc.  All rights reserved.
// A simple UDP source, where every UDP payload is a complete frame
// Implementation

#include "BasicUDPSource.hh"
#include <GroupsockHelper.hh>

//mtk94097 CLI
#include <time.h>
#if 1// add get HW time
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

//#include "x_timer.h"
//#include "x_assert.h"

#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/file.h>

typedef struct _HAL_TIME
{
    unsigned long  u4Seconds;
    unsigned long  u4Micros;
} HAL_TIME_T;

#define IOCTL_BSP_HAL_GET_TIME _IOR('b', 1, HAL_TIME_T)
#define DEV_BSP "/dev/bsp"
void HAL_GetTime(HAL_TIME_T* pTime)
{
    int ret, fd_bsp;

    fd_bsp = open(DEV_BSP, O_RDWR);
    if (fd_bsp < 0)
    {
        printf("[rcS-time] open error\n");
    }

    ret = ioctl(fd_bsp, IOCTL_BSP_HAL_GET_TIME, pTime);
    if (ret < 0)
    {
        printf("[rcS-time] ioctl error\n");
    }

    ret = close(fd_bsp);
    if (ret < 0)
    {
        printf("[rcS-time] close error\n");
    }

}
#endif
#if 1
//For qian's new request: printf direct
typedef struct
{
    int sequence_num;
    unsigned long timestamp_in_header;
    double local_time;
} REC_DATA_T;

REC_DATA_T *prev_receive_time=NULL;
REC_DATA_T *curr_receive_time=NULL;
//For qian's new request: printf direct end
#endif

bool fHavepacketlost = False;
bool bHaveSetMbit = False;

#if 0
bool Log_flag =true;
int count_char =0;

typedef struct
{
    int sequence_num;
    unsigned long header_time;
    unsigned long local_rec_time_sec;
    unsigned long local_rec_time_usec;
} RECEIVE_DATA_T;

RECEIVE_DATA_T  receive_array[10000];


//mtk94097 Timestamp print
#if 0
typedef struct
{
    int start_sequence_num;
    int end_sequence_num;
    unsigned long start_header_time;
    unsigned long end_header_time;
    double total_time;
    unsigned long total_sequence_count;
    double average_header_time;
} STATISTIC_DATA_T;

bool first_in=true;
STATISTIC_DATA_T tmp_data;
#endif
//mtk94097 Timestamp print

#endif
//mtk94097 CLI
BasicUDPSource* BasicUDPSource::createNew(UsageEnvironment& env,
        Groupsock* inputGS)
{
    return new BasicUDPSource(env, inputGS);
}

BasicUDPSource::BasicUDPSource(UsageEnvironment& env, Groupsock* inputGS)
    : FramedSource(env), fInputGS(inputGS), fHaveStartedReading(False),
      fHaveSeenFirstPacket(False)
{
    // Try to use a large receive buffer (in the OS):
    increaseReceiveBufferTo(env, inputGS->socketNum(), 50*1024);

    // Make the socket non-blocking, even though it will be read from only asynchronously, when packets arrive.
    // The reason for this is that, in some OSs, reads on a blocking socket can (allegedly) sometimes block,
    // even if the socket was previously reported (e.g., by "select()") as having data available.
    // (This can supposedly happen if the UDP checksum fails, for example.)
    makeSocketNonBlocking(fInputGS->socketNum());
}

BasicUDPSource::~BasicUDPSource()
{
    envir().taskScheduler().turnOffBackgroundReadHandling(fInputGS->socketNum());
}

void BasicUDPSource::doGetNextFrame()
{
    if (!fHaveStartedReading)
    {
        // Await incoming packets:
        envir().taskScheduler().turnOnBackgroundReadHandling(fInputGS->socketNum(),
                (TaskScheduler::BackgroundHandlerProc*)&incomingPacketHandler, this);
        fHaveStartedReading = True;
    }
}

void BasicUDPSource::doStopGettingFrames()
{
    envir().taskScheduler().turnOffBackgroundReadHandling(fInputGS->socketNum());
    fHaveStartedReading = False;
}


void BasicUDPSource::incomingPacketHandler(BasicUDPSource* source, int /*mask*/)
{
    source->incomingPacketHandler1();
}

void BasicUDPSource::incomingPacketHandler1()
{
    if (!isCurrentlyAwaitingData()) return; // we're not ready for the data yet

    // Read the packet into our desired destination:
    struct sockaddr_in fromAddress;
    unsigned char arHeader[RTP_HEADER_LEN] = {0,};
    unsigned char* fTmpBuffer = fTo;
#if 1//add get HW time
    HAL_TIME_T tm;
#endif
#ifdef MTK_WFD_RTSP
    fFrameSize = fReadCount = 0;
    do
    {
#ifdef NEED_BUFFER_POOL
        if (!fInputGS->handleRead(fReadBuffer, MAX_MTU, fReadSize, fromAddress))
            return;
#else
        if (!fInputGS->handleRead(fTmpBuffer, fMaxSize - fFrameSize, fReadSize, fromAddress))
            return;
#endif

        //printf("fReadSize = %u\n", fReadSize);

        if (fReadCount == 0 && fReadSize == 0)
            return;

        if (fReadSize == 0)
        {
            //printf("read 0, fReadCount = %u, fFrameSize = %u\n", fReadCount, fFrameSize);
            break;
        }

#ifdef NEED_BUFFER_POOL
        unsigned short fSeqNo = ntohs(*(uint16_t*)(fReadBuffer + 2));
        unsigned short fHead = ntohs(*(uint16_t*)(fReadBuffer));

        //printf("fSeqNo = %u\n", fSeqNo);
#else
        unsigned short fSeqNo = ntohs(*(uint16_t*)(fTmpBuffer + 2));
        unsigned short fHead = ntohs(*(uint16_t*)(fTmpBuffer));
#endif

        //mtk94097 CLI

        //For qian's new request: printf direct
#if 1
        extern int user_delay;
#ifdef NEED_BUFFER_POOL
        unsigned long fTimeStamp32 = ntohl(*(uint32_t*)(fReadBuffer + 4));
#else
        unsigned long fTimeStamp32 = ntohl(*(uint32_t*)(fTmpBuffer + 4));
#endif

        struct timeval  local_time_print;
        gettimeofday(&local_time_print,NULL);
        double stardard_time= (double)user_delay/1000.0;

        if (NULL==prev_receive_time)
        {
            printf("[mtk94097] DONGLE_WFD [prev_ts:curr_ts]  [prev_sq:prev_sq]  [delay( curr_time-prev_time) > Stardand(%f)]\n",stardard_time);
            prev_receive_time = (REC_DATA_T *)malloc(sizeof(REC_DATA_T));
            curr_receive_time = (REC_DATA_T *)malloc(sizeof(REC_DATA_T));;
            prev_receive_time->local_time=local_time_print.tv_sec + (double)local_time_print.tv_usec/1000000.0;;
            prev_receive_time->sequence_num=fSeqNo;
            prev_receive_time->timestamp_in_header=fTimeStamp32;
        }

        curr_receive_time->local_time=local_time_print.tv_sec + (double)local_time_print.tv_usec/1000000.0;;
        curr_receive_time->sequence_num=fSeqNo;
        curr_receive_time->timestamp_in_header=fTimeStamp32;

        double diff_time=curr_receive_time->local_time - prev_receive_time->local_time;

        if (diff_time > stardard_time)
            printf("[mtk94097]DONGLE_WFD [%ld:%ld][%d:%d][%f(%f-%f)]\n",
                   prev_receive_time->timestamp_in_header,
                   curr_receive_time->timestamp_in_header,
                   prev_receive_time->sequence_num,
                   curr_receive_time->sequence_num,
                   diff_time,
                   curr_receive_time->local_time,
                   prev_receive_time->local_time);

        prev_receive_time->local_time          =curr_receive_time->local_time;
        prev_receive_time->sequence_num        =curr_receive_time->sequence_num;
        prev_receive_time->timestamp_in_header =curr_receive_time->timestamp_in_header;
#if 0//add get hw time
        HAL_GetTime(&tm);
        printf("[yz][APP] pts:%ld,time:%d.%d\n",curr_receive_time->timestamp_in_header,tm.u4Seconds,tm.u4Micros);
#endif

#endif
        //For qian's new request END

#if 0
#define DEFAULT_DUMP_PATH "/mnt/sda1/Timestamp_Dump.log"
        extern int ts_enable_dbg;

        if (1==ts_enable_dbg)
        {
            if (Log_flag)
            {
                char cmds[128];
                memset(cmds,0,sizeof(cmds));
                memset(receive_array, 0, sizeof(RECEIVE_DATA_T) * 10000);
                sprintf(cmds, "echo 'No.  SeqNo   Header_Time   Local_Time(sec:usec)' >"DEFAULT_DUMP_PATH);
                system(cmds);
                Log_flag =false;
                fprintf(stderr, "[mtk94097]Log_flag == True\n");
            }


            {
                struct timeval  local_time;
#ifdef NEED_BUFFER_POOL
                unsigned long fTimeStamp32 = ntohl(*(uint32_t*)(fReadBuffer + 4));
#else
                unsigned long fTimeStamp32 = ntohl(*(uint32_t*)(fTmpBuffer + 4));
#endif
                gettimeofday(&local_time,NULL);

                receive_array[count_char].sequence_num  = fSeqNo;
                receive_array[count_char].header_time   = fTimeStamp32;
                receive_array[count_char].local_rec_time_sec= local_time.tv_sec;
                receive_array[count_char].local_rec_time_usec=local_time.tv_usec;
                count_char++;

                if (count_char == 10000)
                {
                    FILE* log_fd = fopen(DEFAULT_DUMP_PATH, "a+");
                    if (log_fd == NULL)
                    {
                        fprintf(stderr, "Open "  DEFAULT_DUMP_PATH " fail\n");
                        ts_enable_dbg =0;
                        fprintf(stderr, "[mtk94097]set ts_enable_dbg = 0 to resolve this Error\n");
                    }
                    else
                    {
                        char temp_buf[256];
                        for (int count = 0; count < 10000; ++count)
                        {
                            memset(temp_buf, 0, 256);
                            sprintf(temp_buf, "%d\t%u\t%ld\t%ld:%ld\n", count + 1,
                                    receive_array[count].sequence_num, receive_array[count].header_time,
                                    receive_array[count].local_rec_time_sec,receive_array[count].local_rec_time_usec);
                            fwrite(temp_buf, strlen(temp_buf), 1, log_fd);
                        }

                        fclose(log_fd);
                        sync();
                    }
                    count_char=0;
                    memset(receive_array, 0, sizeof(RECEIVE_DATA_T) * 10000);
                }

            }
        }
#endif
        //mtk94097 CLI

        //mtk94097 Timestamp print
#if 0
        if (first_in)
        {
            memset(&tmp_data, 0, sizeof(STATISTIC_DATA_T));
            first_in=false;
        }
#ifdef NEED_BUFFER_POOL
        unsigned long fTimeStamp32 = ntohl(*(uint32_t*)(fReadBuffer + 4));
#else
        unsigned long fTimeStamp32 = ntohl(*(uint32_t*)(fTmpBuffer + 4));
#endif


        struct timeval  local_time_print;
        gettimeofday(&local_time_print,NULL);
        tmp_data.total_time += local_time_print.tv_sec + (double)local_time_print.tv_usec/1000000.0;
        tmp_data.total_sequence_count ++;

        //start part
        if (fSeqNo%1000 == 1)
        {
            tmp_data.start_sequence_num = fSeqNo;
            tmp_data.start_header_time = fTimeStamp32;
        }
        //end part
        if (fSeqNo%1000 == 0)
        {

            tmp_data.end_sequence_num=fSeqNo;
            tmp_data.end_header_time= fTimeStamp32;
            tmp_data.average_header_time= ((tmp_data.total_time)/(tmp_data.total_sequence_count));
            first_in=true;
            printf("[start_seq= %d] \t[start_headertime= %ld] \t[end_seq= %d] \t[end_headertime= %ld] \t[average= %f(%f/%ld)]\n",
                   tmp_data.start_sequence_num,
                   tmp_data.start_header_time,
                   tmp_data.end_sequence_num,
                   tmp_data.end_header_time,
                   tmp_data.average_header_time,
                   tmp_data.total_time,
                   tmp_data.total_sequence_count
                  );
        }
#endif
        //mtk94097 Timestamp print
        if (!fHaveSeenFirstPacket)
        {
            fNextExpectedSeqNo = fSeqNo + 1;
            fHaveSeenFirstPacket = True;
        }
        else
        {
            if (fNextExpectedSeqNo != fSeqNo)
            {
                int diff = (fSeqNo - fNextExpectedSeqNo + USHORT_MAX - 1) % (USHORT_MAX -1);
				if (diff > 60000)  //assume this packet is wrong order
				{
					printf("Wrong order, Recv SeqNo = %u, Expect SeqNo = %u\n",
						fSeqNo, fNextExpectedSeqNo);
					continue;
				}
                printf("Recv SeqNo = %u, Expect SeqNo = %u, Packet lost = %d\n",
                       fSeqNo, fNextExpectedSeqNo, diff);
                fHavepacketlost = True;
            }
            fNextExpectedSeqNo = fSeqNo + 1;
        }

#ifdef NEED_BUFFER_POOL
        memcpy(fTo + fFrameSize, fReadBuffer + RTP_HEADER_LEN, fReadSize - RTP_HEADER_LEN);
        fFrameSize += (fReadSize - RTP_HEADER_LEN);
        //printf("fFrameSize = %u\n", fFrameSize);
#else
        *(rtp_header_len_t*)fTmpBuffer = *(rtp_header_len_t*)arHeader;
        fTmpBuffer += (fReadSize - RTP_HEADER_LEN);
        fFrameSize += (fReadSize - RTP_HEADER_LEN);
        *(rtp_header_len_t*)arHeader = *(rtp_header_len_t*)fTmpBuffer;
#endif
        // printf("RTP Head 16bit:0x%x \n",fHead);
#if 0
		if (fHead & 0x0080)
        {
            // printf("For WCP2 Debug:%u \n",fHead);
            bHaveSetMbit = True;
            break;
        }
#endif
        if (++fReadCount > MAX_READ_COUNT)
        {
            //printf("read 10ts, fReadCount = %u, fFrameSize = %u\n", fReadCount, fFrameSize);
            break;
        }


    }
    while (1);
#else
    if (!fInputGS->handleRead(fTo, fMaxSize, fFrameSize, fromAddress)) return;

    //Add for 8552
    if (fFrameSize == 0)
    {
        printf("incomingPacketHandler1 receive zero packet\n");
        return;
    }
#endif

    // Tell our client that we have new data:
    afterGetting(this); // we're preceded by a net read; no infinite recursion
}
